# -*- coding: utf-8 -*-
#
# Author: John R. Lenton <john.lenton@canonical.com>
# Author: Natalia B. Bidart <natalia.bidart@canonical.com>
#
# Copyright 2009 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the action queue module."""

from __future__ import with_statement

import logging
import shutil
import unittest
import uuid

import dbus
from dbus.mainloop.glib import DBusGMainLoop
from twisted.internet import defer
from twisted.python.failure import DefaultException, Failure

from contrib.testing.testcase import (
    BaseTwistedTestCase, FakeMain,
    MementoHandler,
)

from ubuntuone.storageprotocol import client, volumes
from ubuntuone.syncdaemon.dbus_interface import DBusInterface
from ubuntuone.syncdaemon.main import Main
from ubuntuone.syncdaemon.action_queue import (
    ActionQueue, ActionQueueCommand, CreateUDF, DeleteVolume, ListVolumes,
    NoisyRequestQueue, RequestQueue
)
from ubuntuone.syncdaemon.event_queue import EventQueue, EVENTS
from ubuntuone.syncdaemon.volume_manager import UDF


DBusInterface.test = True

PATH = u'~/Documents/pdfs/moño/'
NAME = u'UDF-me'
VOLUME = uuid.UUID('12345678-1234-1234-1234-123456789abc')
NODE = uuid.UUID('FEDCBA98-7654-3211-2345-6789ABCDEF12')
USER = u'Dude'


class BasicTestCase(BaseTwistedTestCase):
    """Basic tests to check ActionQueue."""

    def setUp(self):
        """
        prepare to run the test
        """
        BaseTwistedTestCase.setUp(self)
        self.root = self.mktemp('root')
        self.shares = self.mktemp('shares')
        self.data = self.mktemp('data')
        self.partials = self.mktemp('partials')
        self.handler = MementoHandler()
        self.handler.setLevel(logging.ERROR)
        self.main = Main(root_dir=self.root,
                         shares_dir=self.shares,
                         data_dir=self.data,
                         partials_dir=self.partials,
                         host='localhost', port=0,
                         dns_srv=False, ssl=False,
                         disable_ssl_verify=True,
                         realm='fake.realm',
                         mark_interval=60,
                         handshake_timeout=2,
                         glib_loop=DBusGMainLoop(set_as_default=True))
        logging.getLogger('ubuntuone.SyncDaemon').addHandler(self.handler)
        dbus.service.BusName.__del__ = lambda _: None

    def tearDown(self):
        """
        cleanup after the test
        """
        self.main.shutdown()
        shutil.rmtree(self.root)
        shutil.rmtree(self.shares)
        shutil.rmtree(self.data)
        shutil.rmtree(self.partials)
        for record in self.handler.records:
            exc_info = getattr(record, 'exc_info', None)
            if exc_info is not None:
                raise exc_info[0], exc_info[1], exc_info[2]
        BaseTwistedTestCase.tearDown(self)

    def test_creation_requires_main(self):
        """Main instance is needed at creation time."""
        aq = ActionQueue(event_queue=FakedEventQueue(), main=self.main,
                         host='localhost', port=0, dns_srv=False)
        self.assertEquals(self.main, aq.main)

    @defer.inlineCallbacks
    def test_content_queue_has_only_one_op_per_node(self):
        """
        Check that the content queue uniquifies operations per node.
        """
        yield self.main.start()
        # totally fake, we don't care: the messages are only validated on run
        self.main.action_q.download('foo', 'bar', 0, 0)
        self.main.action_q.upload('foo', 'bar', 0, 0, 0, 0, 0)
        self.assertEqual(len(self.main.action_q.content_queue.waiting), 1)

    @defer.inlineCallbacks
    def test_content_queue_has_only_one_op_per_node_even_counting_markers(self):
        """
        Check that the content queue uniquifies operations per node
        even when some of the operations were added using markers.
        """
        yield self.main.start()
        self.main.action_q.download('foo', 'bar', 0, 0)
        self.main.action_q.uuid_map.set('foo', 'feh')
        self.main.action_q.uuid_map.set('bar', 'bah')
        self.main.action_q.upload('feh', 'bah', 0, 0, 0, 0, 0)
        self.assertEqual(len(self.main.action_q.content_queue.waiting), 1)

    @defer.inlineCallbacks
    def test_aq_resolve_uuid_maybe(self):
        """
        Check action_q.resolve_uuid_maybe does what it's supposed to
        """
        yield self.main.start()
        self.assertEqual(self.main.action_q.resolve_uuid_maybe('foo'), 'foo')
        self.main.action_q.uuid_map.set('foo', 'feh')
        self.assertEqual(self.main.action_q.resolve_uuid_maybe('foo'), 'feh')


class TestNoisyRQ(unittest.TestCase):
    """
    Tests for NoisyRequestQueue
    """

    def test_noisy_rq_blurts_about_head(self):
        """
        Test NRQ calls its callback when head is set
        """
        rq = NoisyRequestQueue('name', None,
                               lambda h, w:
                                   setattr(self, 'result', (h, tuple(w))))
        rq._head = 'blah'
        self.assertEqual(self.result, ('blah', ()))

    def test_noisy_rq_blurts_about_waiting(self):
        """
        Test NRQ calls its callback when the waiting queue is altered.
        """
        class BlackHole(object):
            """The universal tool."""
            __call__ = __getattr__ = lambda *_, **__: BlackHole()
        class FakeCommand(object):
            """Yet another fake action queue command"""
            def run(self):
                """run that just succeeds"""
                return defer.succeed(None)
            def is_runnable(self):
                """returns True"""
                return True
        def cb(head, waiting):
            """NRQ testing callback"""
            evts.append((head, tuple(waiting)))
        evts = []
        cmd = FakeCommand()
        cmd2 = FakeCommand()
        rq = NoisyRequestQueue('name', BlackHole(), cb)
        rq.queue(cmd2)
        rq.queue_top(cmd)
        rq.run()
        self.assertEqual(evts, [(None, ()),           # __init__
                                (None, (cmd2,)),      # queue
                                (None, (cmd, cmd2)),  # queue_top
                                (cmd, (cmd2,)),       # run
                                (None, (cmd2,)),      # done
                                ])


def noop(*args, **kwargs):
    """No op."""


class DummyClass(object):
    """Dummy class, does nothing."""

    def __getattribute__(self, name):
        """Any attribute is a no-op."""
        return noop


class FakedTwistedTestCase(unittest.TestCase):
    """Helper for by-pass Twisted."""

    def setUp(self):
        """Init."""
        self.factory = ActionQueue(event_queue=FakedEventQueue(), main=None,
                                   host='localhost', port=0, dns_srv=False)
        self.factory.connector = DummyClass()

        self.client = self.factory.protocol()
        self.client.factory = self.factory
        self.client.transport = DummyClass()
        self.client.connectionMade() # simulate connection

        self.factory.event_queue.events = [] # reset events

    def tearDown(self):
        """Clean up."""
        self.factory = None


class FakedEventQueue(EventQueue):
    """Faked event queue."""

    def __init__(self):
        """Initialize a faked event queue."""
        # pylint: disable-msg=W0231
        self._listeners = []
        self.events = []

    def push(self, event_name, *args, **kwargs):
        """Faked event pushing."""
        self.events.append((event_name, args, kwargs))


class FakedVolume(object):
    """Faked volume."""


class ConnectionTestCase(FakedTwistedTestCase):
    """Test TCP/SSL connection mechanism for ActionQueue."""

    def assert_connection_state_reset(self, factory):
        """Test connection state is properly reset."""
        self.assertTrue(factory.client is None)
        self.assertTrue(factory.connector is None)

    def test_init(self):
        """Test connection init state."""
        factory = ActionQueue(event_queue=FakedEventQueue(), main=None,
                              host='localhost', port=0, dns_srv=False)
        self.assert_connection_state_reset(factory)

    def test_connector_gets_assigned_on_connect(self):
        """Test factory's connector gets assigned on connect."""
        self.assertTrue(self.factory.connector is not None)


class VolumeManagementTestCase(FakedTwistedTestCase):
    """Test Volume managemenr for ActionQueue."""

    def test_callbacks_are_connected(self):
        """Test callbacks are connected."""
        # pylint: disable-msg=W0212
        self.assertEquals(self.factory.client._volume_created_callback,
                          self.factory._volume_created_callback)
        self.assertEquals(self.factory.client._volume_deleted_callback,
                          self.factory._volume_deleted_callback)

    def test_volume_created_push_event(self):
        """Volume created callback push proper event."""
        volume = FakedVolume()
        self.factory._volume_created_callback(volume)
        self.assertEquals([('SV_VOLUME_CREATED', (), {'volume': volume})],
                          self.factory.event_queue.events)

    def test_volume_deleted_push_event(self):
        """Volume deleted callback push proper event."""
        volume_id = VOLUME
        self.factory._volume_deleted_callback(volume_id)
        self.assertEquals([('SV_VOLUME_DELETED', (), {'volume_id': volume_id})],
                          self.factory.event_queue.events)

    def test_valid_events(self):
        """Volume events are valid in EventQueue."""
        new_events = ('SV_VOLUME_CREATED', 'SV_VOLUME_DELETED',
                      'AQ_CREATE_UDF_OK', 'AQ_CREATE_UDF_ERROR',
                      'AQ_LIST_VOLUMES', 'AQ_LIST_VOLUMES_ERROR',
                      'AQ_DELETE_VOLUME_OK', 'AQ_DELETE_VOLUME_ERROR')
        for event in new_events:
            self.assertTrue(event in EVENTS)

        self.assertEquals(('volume_id', 'node_id'), EVENTS['SV_VOLUME_CREATED'])
        self.assertEquals(('volume_id'), EVENTS['SV_VOLUME_DELETED'])
        self.assertEquals(('volume_id', 'node_id', 'marker'),
                          EVENTS['AQ_CREATE_UDF_OK'])
        self.assertEquals(('error', 'marker'), EVENTS['AQ_CREATE_UDF_ERROR'])
        self.assertEquals(('volumes',), EVENTS['AQ_LIST_VOLUMES'])
        self.assertEquals(('error',), EVENTS['AQ_LIST_VOLUMES_ERROR'])
        self.assertEquals(('volume_id',), EVENTS['AQ_DELETE_VOLUME_OK'])
        self.assertEquals(('volume_id', 'error',), EVENTS['AQ_DELETE_VOLUME_ERROR'])

    def test_create_udf(self):
        """Test volume creation."""
        path = PATH
        name = NAME
        res = self.factory.create_udf(path, name, marker=None)
        self.assertTrue(res is None) # this is what start returns

    def test_list_volumes(self):
        """Test volume listing."""
        res = self.factory.list_volumes()
        self.assertTrue(res is None) # this is what start returns

    def test_delete_volume(self):
        """Test volume deletion."""
        volume_id = VOLUME
        res = self.factory.delete_volume(volume_id)
        self.assertTrue(res is None) # this is what start returns


class CreateUDFTestCase(FakedTwistedTestCase):
    """Test for CreateUDF ActionQueueCommand."""

    def setUp(self):
        """Init."""
        super(CreateUDFTestCase, self).setUp()

        request_queue = RequestQueue(name='foo', action_queue=self.factory)

        self.marker = VOLUME
        self.command = CreateUDF(request_queue, PATH, NAME, marker=self.marker)

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        self.assertTrue(isinstance(self.command, ActionQueueCommand))

    def test_init(self):
        """Test creation."""
        self.assertEquals(PATH, self.command.path)
        self.assertEquals(NAME, self.command.name)
        self.assertEquals(self.marker, self.command.marker)

    def test_run_returns_a_deferred(self):
        """Test a deferred is returned."""
        res = self.command._run()
        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')

    def test_run_calls_protocol(self):
        """Test protocol's create_udf is called."""
        original = self.command.action_queue.client.create_udf
        self.called = False

        def check(path, name):
            """Take control over client's feature."""
            self.called = True
            self.assertEquals(PATH, path)
            self.assertEquals(NAME, name)

        self.command.action_queue.client.create_udf = check

        res = self.command._run()

        self.assertTrue(self.called, 'command was called')

        self.command.action_queue.client.create_udf = original

    def test_handle_success_push_event(self):
        """Test AQ_CREATE_UDF_OK is pushed on success."""
        request = client.CreateUDF(self.factory.client, PATH, NAME)
        request.volume_id = VOLUME
        request.node_id = NODE
        res = self.command.handle_success(success=request)
        events = [('AQ_CREATE_UDF_OK', (), {'volume_id': VOLUME,
                                            'node_id': NODE,
                                            'marker': self.marker})]
        self.assertEquals(events, self.command.action_queue.event_queue.events)
        self.assertEquals(request, res)

    def test_handle_failure_push_event(self):
        """Test AQ_CREATE_UDF_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))
        res = self.command.handle_failure(failure=failure)
        events = [('AQ_CREATE_UDF_ERROR', (),
                    {'error': msg, 'marker': self.marker})]
        self.assertEquals(events, self.command.action_queue.event_queue.events)
        self.assertTrue(res is None)


class ListVolumesTestCase(FakedTwistedTestCase):
    """Test for ListVolumes ActionQueueCommand."""

    def setUp(self):
        """Init."""
        super(ListVolumesTestCase, self).setUp()

        request_queue = RequestQueue(name='foo', action_queue=self.factory)

        self.command = ListVolumes(request_queue)

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        self.assertTrue(isinstance(self.command, ActionQueueCommand))

    def test_run_returns_a_deferred(self):
        """Test a deferred is returned."""
        res = self.command._run()
        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')

    def test_run_calls_protocol(self):
        """Test protocol's list_volumes is called."""
        original = self.command.action_queue.client.list_volumes
        self.called = False

        def check():
            """Take control over client's feature."""
            self.called = True

        self.command.action_queue.client.list_volumes = check

        res = self.command._run()

        self.assertTrue(self.called, 'command was called')

        self.command.action_queue.client.list_volumes = original

    def test_handle_success_push_event(self):
        """Test AQ_LIST_VOLUMES is pushed on success."""
        request = client.ListVolumes(self.factory.client)
        request.volumes = [FakedVolume(), FakedVolume()]
        res = self.command.handle_success(success=request)
        events = [('AQ_LIST_VOLUMES', (), {'volumes': request.volumes})]
        self.assertEquals(events, self.command.action_queue.event_queue.events)
        self.assertEquals(request, res)

    def test_handle_failure_push_event(self):
        """Test AQ_LIST_VOLUMES_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))
        res = self.command.handle_failure(failure=failure)
        events = [('AQ_LIST_VOLUMES_ERROR', (), {'error': msg})]
        self.assertEquals(events, self.command.action_queue.event_queue.events)
        self.assertTrue(res is None)


class DeleteVolumeTestCase(FakedTwistedTestCase):
    """Test for DeleteVolume ActionQueueCommand."""

    def setUp(self):
        """Init."""
        super(DeleteVolumeTestCase, self).setUp()

        request_queue = RequestQueue(name='foo', action_queue=self.factory)

        self.command = DeleteVolume(request_queue, VOLUME)

    def test_is_action_queue_command(self):
        """Test proper inheritance."""
        self.assertTrue(isinstance(self.command, ActionQueueCommand))

    def test_init(self):
        """Test creation."""
        self.assertEquals(VOLUME, self.command.volume_id)

    def test_run_returns_a_deferred(self):
        """Test a deferred is returned."""
        res = self.command._run()
        self.assertTrue(isinstance(res, defer.Deferred), 'deferred returned')

    def test_run_calls_protocol(self):
        """Test protocol's delete_volume is called."""
        original = self.command.action_queue.client.delete_volume
        self.called = False

        def check(volume_id):
            """Take control over client's feature."""
            self.called = True
            self.assertEquals(VOLUME, volume_id)

        self.command.action_queue.client.delete_volume = check

        res = self.command._run()

        self.assertTrue(self.called, 'command was called')

        self.command.action_queue.client.delete_volume = original

    def test_handle_success_push_event(self):
        """Test AQ_DELETE_VOLUME_OK is pushed on success."""
        request = client.DeleteVolume(self.factory.client, volume_id=VOLUME)
        res = self.command.handle_success(success=request)
        events = [('AQ_DELETE_VOLUME_OK', (), {'volume_id': VOLUME})]
        self.assertEquals(events, self.command.action_queue.event_queue.events)
        self.assertEquals(request, res)

    def test_handle_failure_push_event(self):
        """Test AQ_DELETE_VOLUME_ERROR is pushed on failure."""
        msg = 'Something went wrong'
        failure = Failure(DefaultException(msg))
        res = self.command.handle_failure(failure=failure)
        events = [('AQ_DELETE_VOLUME_ERROR', (), {'volume_id': VOLUME, 'error': msg})]
        self.assertEquals(events, self.command.action_queue.event_queue.events)
        self.assertTrue(res is None)


class FilterEventsTestCase(BaseTwistedTestCase):
    """Tests for event filtering when a volume is not of our interest."""

    def setUp(self):
        """Init."""
        BaseTwistedTestCase.setUp(self)
        self.log = logging.getLogger("ubuntuone.SyncDaemon.TEST")
        self.log.info("starting test %s.%s", self.__class__.__name__,
                      self._testMethodName)
        self.root_dir = self.mktemp('root_dir')
        self.data_dir = self.mktemp('data_dir')
        self.shares_dir = self.mktemp('shares_dir')
        self.partials_dir = self.mktemp('partials_dir')
        self.main = FakeMain(self.root_dir, self.shares_dir,
                             self.data_dir, self.partials_dir)
        self.vm = self.main.vm

        self.action_queue = ActionQueue(FakedEventQueue(), self.main,
                                        host='localhost', port=0,
                                        dns_srv=False)

    def tearDown(self):
        """Clean up."""
        self.action_queue = None

        self.main.shutdown()
        self.rmtree(self.root_dir)
        self.rmtree(self.data_dir)
        self.rmtree(self.shares_dir)
        self.log.info("finished test %s.%s", self.__class__.__name__,
                      self._testMethodName)
        return BaseTwistedTestCase.tearDown(self)

    @defer.inlineCallbacks
    def test_SV_HASH_NEW_is_pushed_for_subscrined_volume(self):
        """SV_HASH_NEW is filtered when the volume is unsubscribed."""
        udf_id = 'udf_id'
        udf_volume = volumes.UDFVolume(udf_id, 'udf_node',
                                       u'~/ñoño'.encode("utf8"))
        path = self.vm._build_udf_path(udf_volume.suggested_path)
        udf = UDF.from_udf_volume(udf_volume, path)
        yield self.vm.add_udf(udf)
        yield self.vm.subscribe_udf(udf_id)
        assert self.vm.udfs[udf_id].subscribed

        kwargs = dict(share_id=udf_id, node_id=None, hash=None)
        self.action_queue._node_state_callback(**kwargs)
        self.assertEquals([('SV_HASH_NEW', (), kwargs)],
                           self.action_queue.event_queue.events)

    @defer.inlineCallbacks
    def test_SV_HASH_NEW_is_filtered_for_unsubscrined_volume(self):
        """SV_HASH_NEW is filtered when the volume is unsubscribed."""
        # build a VM and add it an UDF with subscribed to False
        udf_id = 'udf_id'
        udf_volume = volumes.UDFVolume(udf_id, 'udf_node',
                                       u'~/ñoño'.encode("utf8"))
        path = self.vm._build_udf_path(udf_volume.suggested_path)
        udf = UDF.from_udf_volume(udf_volume, path)
        yield self.vm.add_udf(udf)
        yield self.vm.unsubscribe_udf(udf_id)
        assert not self.vm.udfs[udf_id].subscribed

        self.action_queue._node_state_callback(share_id=udf_id,
                                               node_id=None, hash=None)
        self.assertEquals([], self.action_queue.event_queue.events)

    def test_SV_HASH_NEW_doesnt_fail_for_non_udf(self):
        """SV_HASH_NEW keeps working like before for non-udfs."""
        other_id = 'not in udfs'
        assert other_id not in self.vm.udfs

        kwargs = dict(share_id=other_id, node_id=None, hash=None)
        self.action_queue._node_state_callback(**kwargs)
        self.assertEquals([('SV_HASH_NEW', (), kwargs)],
                          self.action_queue.event_queue.events)
